#ifndef SCREEM_DTD_H
#define SCREEM_DTD_H

#include <glib-object.h>
#include <gtk/gtkwidget.h>

#include <libxml/tree.h>
#include <libxml/parser.h>

#define SCREEM_TYPE_DTD           (screem_dtd_get_type ())
#define SCREEM_DTD(obj)           (G_TYPE_CHECK_INSTANCE_CAST ((obj), SCREEM_TYPE_DTD, ScreemDTD))
#define SCREEM_IS_DTD(obj)        (G_TYPE_CHECK_INSTANCE_TYPE ((obj), SCREEM_TYPE_DTD))

				   
typedef enum ScreemDTDClose {
	SHOULD,
	MUST,
	MUSTNT
} ScreemDTDClose;

typedef struct ScreemDTD ScreemDTD;

typedef struct ScreemDTDPrivate ScreemDTDPrivate;
typedef struct ScreemDTDElement ScreemDTDElement;
typedef struct ScreemDTDAttribute ScreemDTDAttribute;

struct ScreemDTD {
	GObject parent_object;

	ScreemDTDPrivate *private;
};

typedef struct {
	GObjectClass parent_class;

	void (*load_module)( ScreemDTD *dtd, const gchar *element,
			     gpointer *module );

} ScreemDTDClass;

GType screem_dtd_get_type();		
ScreemDTD *screem_dtd_new();
ScreemDTD *screem_dtd_new_with_id( const gchar *id );
gboolean screem_dtd_load( ScreemDTD *dtd, const gchar *uri, 
			  ScreemDTD *parent, gboolean online );
ScreemDTDElement *screem_dtd_valid_element( ScreemDTD *dtd, 
					    const gchar *element );
ScreemDTDAttribute *screem_dtd_valid_attr( ScreemDTD *dtd,
					   const gchar *element,
					   const gchar *attr );
gboolean screem_dtd_valid_child( ScreemDTD *dtd,
				 const gchar *element,
				 const gchar *child );
ScreemDTDClose screem_dtd_element_close_state( ScreemDTD *dtd, 
					       const gchar *element );


GtkWidget *screem_dtd_attribute_menu( ScreemDTD *dtd, const gchar *element,
				      gpointer data );
GtkWidget *screem_dtd_allowed_menu( ScreemDTD *dtd, const gchar *element );

gboolean screem_dtd_is_root_element( ScreemDTD *dtd, const gchar *name );

const gchar *screem_dtd_root_element_name( ScreemDTD *dtd );

gboolean screem_dtd_parse_doctype( const gchar *element,
				   gchar **id,
				   gchar **system );

gboolean screem_dtd_attr_is_uri( ScreemDTD *dtd,
				 const gchar *element, const gchar *attr );

/* ScreemDTDElement access functions */
const GSList *screem_dtd_get_elements( ScreemDTD *dtd );
const gchar *screem_dtd_element_get_name( const ScreemDTDElement *element );
const GSList *screem_dtd_element_get_attrs( const ScreemDTDElement *element );

GSList *screem_dtd_get_element_names( ScreemDTD *dtd );

/* ScreemDTDAttribute access functions */
const gchar *
screem_dtd_attribute_get_name( const ScreemDTDAttribute *attribute );
const GSList *
screem_dtd_attribute_get_values( const ScreemDTDAttribute *attribute );

GSList *screem_dtd_get_attribute_names( ScreemDTD *dtd, const gchar *element );

/* for debuging purposes */
void screem_dtd_screen_dump( ScreemDTD *dtd );

#endif
