/*  Screem:  screem-print.c,
 * 
 *  Copyright (C) 2002  David A Knight
 *
 *  The code in this file is derived from the code in gedit-print.c of
 *  Gedit 2, which is Copyright (C) 2000, 2001 Chema Celorio, Paolo Maggi
 *  and the gedit Team 1998-2001. See the Gedit AUTHORS file for a list
 *  of people on the gedit Team.
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 *  For contact information with the author of this source code please see
 *  the AUTHORS file.  If there is no AUTHORS file present then check the
 *  about box under the help menu for a contact address
 */

#include <gconf/gconf-client.h>

#include "screem-print.h"

#include "screem-page.h"

static GnomePrintConfig *screem_print_config = NULL;

#define A4_WIDTH (210.0 * 72 / 25.4)
#define A4_HEIGHT (297.0 * 72 / 25.4)

ScreemPrintJob* screem_print_job_new( ScreemPage *page, gboolean preview )
{
	GConfClient *client;
	gchar *font;
	ScreemPrintJob *job;

	if( ! screem_print_config  ) {
		screem_print_config = gnome_print_config_default();
		g_return_val_if_fail( screem_print_config != NULL, NULL );
		/*		gnome_print_config_set( screem_print_config, "Settings.Transport.Backend", "lpr" );
				gnome_print_config_set( screem_print_config, "Printer", "GENERIC" );*/
	}

	client = gconf_client_get_default();
	gconf_client_add_dir( client, "/apps/screem/editor/print",
			      GCONF_CLIENT_PRELOAD_ONELEVEL, NULL );

	job = g_new0( ScreemPrintJob, 1 );

	job->page = page;
	job->preview = preview;

	job->range_type = GNOME_PRINT_RANGE_ALL;

	job->page_num = 0;

	job->page_width = A4_WIDTH;
	job->page_height = A4_HEIGHT;
	job->margin_top = CM( 1 );
	job->margin_left = CM( 1 );
	job->margin_right = CM( 1 );
	job->margin_bottom = CM( 1 );
	job->header_height = CM( 0 );

	font = gconf_client_get_string( client, 
					"/apps/screem/editor/print/body_font",
					NULL );
	job->font_body = gnome_font_find_closest_from_full_name( font );
	g_free( font );

	font = gconf_client_get_string(client, 
				       "/apps/screem/editor/print/header_font",
				       NULL );
	job->font_header = gnome_font_find_closest_from_full_name( font );
	g_free( font );

	font = gconf_client_get_string( client, 
					"/apps/screem/editor/print/lines_font",
					NULL );
	job->font_numbers = gnome_font_find_closest_from_full_name( font );
	g_free( font );


	job->print_header = 
		gconf_client_get_bool( client,
				       "/apps/screem/editor/print/headers",
				       NULL );
	job->print_wrap_lines =
		gconf_client_get_bool( client,
				       "/apps/screem/editor/print/weap_lines",
				       NULL );
	job->print_line_numbers = 
		gconf_client_get_int( client,
				       "/apps/screem/editor/print/nlines",
				       NULL );
	g_object_unref( client );

	if((! job->font_body) || (!job->font_header) || (!job->font_numbers)) {
		screem_print_job_destroy( job );
		job = NULL;
	} else {
		job->header_height = 2.5 * gnome_font_get_size(job->font_header);
		job->numbers_width = CM( 0 );

		job->first_line_to_print = 1;
		job->printed_lines = 0;
	}


	job->config = screem_print_config;
	gnome_print_config_ref( job->config );

	return job;
}

void screem_print_job_destroy( ScreemPrintJob *job )
{
	if( job->config )
		gnome_print_config_unref( job->config );

	if( job->print_ctx )
		g_object_unref( job->print_ctx );

	if( job->print_job )
		g_object_unref( job->print_job );

	g_free( job );
}
